# RevKit: A Toolkit for Reversible Circuit Design (www.revkit.org)
# Copyright (C) 2009-2011  The RevKit Developers <revkit@informatik.uni-bremen.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from core.BaseItem import *

import glob
import os

from revkit import circuit, binary_truth_table, read_realization, read_specification

from ui.DesignerWidget import *
from ui.PathBenchmarks import *

class PathBenchmarks( DesignerWidget ):
    filenames = None
    fileTypeChanged = pyqtSignal( QString )
    fileNamesChanged = pyqtSignal()
    parent = None

    def __init__( self, parent = None ):
        DesignerWidget.__init__( self, Ui_PathBenchmarks, parent )

        # Completer
        completer = QCompleter( self )
        completer.setModel( QDirModel( completer ) )
        self.path.setCompleter( completer )

        # Select Path
        self.connect( self.path, SIGNAL( 'textChanged(const QString&)' ), self.path_changed )
        self.connect( self.select_path, SIGNAL( 'clicked()' ), self.select_path_clicked )

        # Select File Type
        self.connect( self.filetype, SIGNAL( 'currentIndexChanged(int)' ), self.update )
        self.connect( self.filetype, SIGNAL( 'currentIndexChanged(int)' ), self.slot_file_type_changed )

    def path_changed( self, text ):
        if len( self.parent.scene().graph.outEdges( self.parent.providesPorts[0] ) ) > 0:
            self.update()
            return

        # Guess filetype
        path = str( self.path.text() )
        if os.path.exists( path ):
            extensions = [ "*.pla", "*.real", "*.spec" ]
            counts = map( len, map( lambda ext: glob.glob( os.path.join( path, ext ) ), extensions ) )
            if counts.count( max( counts ) ) == 1:
                self.filetype.setCurrentIndex( counts.index( max( counts ) ) )
        self.update()

    def select_path_clicked( self ):
        self.path.setText( QFileDialog.getExistingDirectory( self, "Open Directory" ) )

    def update( self ):
        path = str( self.path.text() )
        if os.path.exists( path ):
            extensions = [ "*.pla", "*.real", "*.spec" ]
            names = [ "function", "circuit", "truth table" ]
            self.filenames = glob.glob( os.path.join( path, extensions[self.filetype.currentIndex()] ) )
            self.info.setText( "Found %d %s%s" % ( len( self.filenames ),
                                                   names[self.filetype.currentIndex()],
                                                   "s" if len( self.filenames ) != 1 else "" ) )
        else:
            self.filenames = None
            self.info.setText( "Please choose a location" )

        self.fileNamesChanged.emit()

    def slot_file_type_changed( self, index ):
        types = [ "PLA", "Circuit", "Truth Table" ]
        self.fileTypeChanged.emit( types[index] )

class PathBenchmarksItem( BaseItem ):
    "This item opens a set of benchmarks (either Boolean functions provided in *.pla-files, Truth Tables provided in *.spec-files, or Circuits provided in *.real-files) and separately pass them to the succeeding items. It can be utilized, if a design process should be applied to a larger set of benchmarks. A right-click on the item opens a file browser, where a path including the respective files can be selected. When enlarging the item, the type of benchmark to consider can be selected."
    description = "Path Benchmarks"
    custom_size = QSize( 320, 150 )
    iconname = "document-open-folder"
    currentIndex = 0

    def __init__( self, parent = None ):
        BaseItem.__init__( self, parent )

        self.provides( [ "PLA" ] )

    def onCreate( self ):
        self.setText( "Functions from path" )
        self.setState( self.UNCONFIGURED )

        self.widget = self.mainWidget( PathBenchmarks() )
        self.widget.parent = self
        self.widget.fileTypeChanged.connect( self.onFileTypeChanged )
        self.widget.fileNamesChanged.connect( self.onFileNamesChanged )

        # Observe port
        self.providesPorts[0].connectionsChanged.connect( self.onConnectionsChanged )

        # Actions
        self.actions = [ [ "document-open-folder", "Choose path", self.widget.select_path_clicked ] ]

    def initialize( self ):
        self.currentIndex = 0

    def executeEvent( self, inputs ):
        index = self.widget.filetype.currentIndex()
        filename = self.widget.filenames[self.currentIndex]
        self.currentIndex += 1
        if index == 0: # Functions
            return [ filename ]
        elif index == 1: # Circuit
            circ = circuit()
            read_realization( circ, filename )
            circ.circuit_name = filename[filename.rfind('/') + 1:filename.rfind('.')]
            return [ circ ]
        elif index == 2: # Specification
            spec = binary_truth_table()
            read_specification( spec, filename )
            return [ spec ]

    def num_runs( self ):
        return len( self.widget.filenames )


    def onFileTypeChanged( self, filetype ):
        self.providesPorts[0].datatype = filetype
        self.providesPorts[0].datatypeWidth = 200
        self.providesPorts[0].update()

        self.setText( "%ss from path" % filetype[:].replace( "PLA", "Function" ) )

    def onFileNamesChanged( self ):
        self.setState( self.CONFIGURED if self.widget.filenames is not None else self.UNCONFIGURED )

    def onConnectionsChanged( self ):
        self.widget.filetype.setEnabled( len( self.scene().graph.outEdges( self.providesPorts[0] ) ) == 0 )

    def setPath( self, v ):
        self.widget.path.setText( v )
        self.widget.update()

    def setFiletype( self, v ):
        self.widget.filetype.setCurrentIndex( int( str( v ) ) )

    conf_path = pyqtProperty( "QString", lambda x: x.widget.path.text(), setPath )
    conf_filetype = pyqtProperty( "int", lambda x: x.widget.filetype.currentIndex(), setFiletype )

